unit XLink;

// XLink Object Model 1.0.0
// Delphi 3, 4, 5, 6, and Kylix Implementation
//
// Copyright (c) 2002 by Dieter Khler
// ("http://www.philo.de/xml/")
//                            
// Definitions:
// - "Package" refers to the collection of files distributed by
//   the Copyright Holder, and derivatives of that collection of
//   files created through textual modification.
// - "Standard Version" refers to such a Package if it has not
//   been modified, or has been modified in accordance with the
//   wishes of the Copyright Holder.
// - "Copyright Holder" is whoever is name in the copyright or
//   copyrights for the package.
// - "You" is you, if you're thinking about copying or distributing
//   this Package.
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Package"), to deal in the Package without restriction,
// including without limitation the rights to use, copy, modify,
// merge, publish, distribute, sublicense, and/or sell copies of the
// Package, and to permit persons to whom the Package is furnished
// to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Package.
//
// You may modify your copy of this Package in any way, provided
// that you insert a prominent notice in each changed file stating
// how and when you changed a file, and provided that you do at
// least one of the following:
//
// a) allow the Copyright Holder to include your modifications in
// the Standard Version of the Package.
//
// b) use the modified Package only within your corporation or
// organization.
//
// c) rename any non standard executables, units, and classes so
// the names do not conflict with standard executables, units, and
// classes, and provide a separate manual page that clearly documents
// how it differs from the standard version.
//
// d) make other distribution arrangements with the Copyright Holder.
//
// The name of the Copyright Holder may not be used to endorse or
// promote products derived from this Package without specific prior
// written permission.
//
// THE PACKAGE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// PACKAGE OR THE USE OR OTHER DEALINGS IN THE PACKAGE.

interface

uses
  Classes, sysutils, contnrs;

type
  EXLinkException = class(Exception);
  EXLink_Hierarchy_Request_Err = class(EXLinkException);
  EXLink_Inuse_Node_Err = class(EXLinkException);
  EXLink_Syntax_Err = class(EXLinkException);

  TXLinkType = (ltNone,
                ltExtended,
                ltLocator,
                ltArc,
                ltResource,
                ltTitle);

  TXLinkShow = (lsUnspecified,
                lsNew,
                lsReplace,
                lsEmbed,
                lsOther,
                lsNone);

  TXLinkActuate = (laUnspecified,
                   laOnLoad,
                   laOnRequest,
                   laOther,
                   laNone);

  TXLinkTypes = set of TXLinkType;

  TXLinkNode = class;
  TXLinkTitle = class;
  TXLinkArc = class;
  TXLinkLocator = class;
  TXLinkResource = class;

  TXLinkList = class
  protected
    FItemTypes: TXLinkTypes;
    FList: TObjectList;
    FOwnsXLinks: boolean;
    function getItems(index: integer): TXLinkNode; virtual;
    function getLength: integer; virtual;
    function getOwnsXLinks: boolean; virtual;
    procedure setOwnsXLinks(value: boolean); virtual;
    function Add(item: TXLinkNode): integer; virtual;
    function AddCopyOf(item: TXLinkNode;
                       deep: boolean): TXLinkNode; virtual;
    function Extract(item: TXLinkNode): TXLinkNode; virtual;
    function FindInstanceOf(AClass: TClass; AExact: boolean = true; AStartAt: integer = 0): integer; virtual;
    procedure Insert(index: integer; item: TXLinkNode); virtual;
    property OwnsXLinks: boolean read getOwnsXLinks write setOwnsXLinks;
  public
    constructor Create(const itemTypes: TXLinkTypes);
    destructor Destroy; override;
    procedure Clear; virtual;
    procedure Delete(index: integer); virtual;
    procedure Exchange(index1, index2: integer); virtual;
    function First: TXLinkNode; virtual;
    function IndexOf(item: TXLinkNode): integer; virtual;
    function Last: TXLinkNode; virtual;
    procedure Move(curIndex, newIndex: integer); virtual;
    function Remove(item: TXLinkNode): integer; virtual;
    property Items[index: integer]: TXLinkNode read getItems;
    property Length: integer read getLength;
  end;

  TXLinkNode = class
  protected
    FAssociatedObject: TObject;
    FOwner: TXLinkList;
    FXLinkType: TXLinkType;
    function getTypeAsWideString: wideString; virtual;
  public
    constructor Create;
    property AssociatedObject: TObject read FAssociatedObject write FAssociatedObject;
    property Owner: TXLinkList read FOwner;
    property XLinkType: TXLinkType read FXLinkType;
    property TypeAsWideString: wideString read getTypeAsWideString;
  end;

  TXLinkExtended = class (TXLinkNode)
  protected
    FArcs: TXLinkList;
    FLocators: TXLinkList;
    FResources: TXLinkList;
    FRole: wideString;
    FTitle: wideString;
    FTitles: TXLinkList;
    function createXLinkArc(arcrole, title: wideString;
                            show: TXLinkShow;
                            actuate: TXLinkActuate;
                            XLinkFrom, XLinkTo: wideString;
                            associatedObject: TObject): TXLinkArc; virtual;
    function createXLinkLocator(href, role, title, XLinklabel: wideString;
                                associatedObject: TObject): TXLinkLocator; virtual;
    function createXLinkResource(role, title, XLinkLabel: wideString;
                                 associatedObject: TObject): TXLinkResource; virtual;
    function createXLinkTitle(associatedObject: TObject): TXLinkTitle; virtual;
  public
    constructor Create;
    destructor Destroy; override;
    procedure AppendContentOf(source: TXLinkExtended;
                              copyObjInf: boolean); virtual;
    function AppendXLinkArc(arcrole, title: wideString;
                            show: TXLinkShow;
                            actuate: TXLinkActuate;
                            XLinkFrom, XLinkTo: wideString;
                            associatedObject: TObject): TXLinkArc; virtual;
    function AppendXLinkLocator(href, role, title, XLinklabel: wideString;
                                associatedObject: TObject): TXLinkLocator; virtual;
    function AppendXLinkResource(role, title, XLinkLabel: wideString;
                                 associatedObject: TObject): TXLinkResource; virtual;
    function AppendXLinkTitle(associatedObject: TObject): TXLinkTitle; virtual;
    function InsertXLinkArc(index: integer;
                            arcrole, title: wideString;
                            show: TXLinkShow;
                            actuate: TXLinkActuate;
                            XLinkFrom, XLinkTo: wideString;
                            associatedObject: TObject): TXLinkArc; virtual;
    function InsertXLinkLocator(index: integer;
                                href, role, title, XLinklabel: wideString;
                                associatedObject: TObject): TXLinkLocator; virtual;
    function InsertXLinkResource(index: integer;
                                 role, title, XLinkLabel: wideString;
                                 associatedObject: TObject): TXLinkResource; virtual;
    function InsertXLinkTitle(index: integer;
                              associatedObject: TObject): TXLinkTitle; virtual;
    property Arcs: TXLinkList read FArcs;
    property Locators: TXLinkList read FLocators;
    property Resources: TXLinkList read FResources;
    property Role: wideString read FRole write FRole;
    property Title: wideString read FTitle write FTitle;
    property Titles: TXLinkList read FTitles;
  end;

  TXLinkLocator = class (TXLinkNode)
  protected
    FHref: wideString;
    FRole: wideString;
    FTitle: wideString;
    FTitles: TXLinkList;
    FXLinkLabel: wideString;
    function createXLinkTitle(associatedObject: TObject): TXLinkTitle; virtual;
  public
    constructor Create;
    destructor Destroy; override;
    procedure AppendContentOf(source: TXLinkLocator;
                              copyObjInf: boolean); virtual;
    function AppendXLinkTitle(associatedObject: TObject): TXLinkTitle; virtual;
    function InsertXLinkTitle(index: integer;
                              associatedObject: TObject): TXLinkTitle; virtual;
    property Href: wideString read FHref write FHref;
    property Role: wideString read FRole write FRole;
    property Title: wideString read FTitle write FTitle;
    property Titles: TXLinkList read FTitles;
    property XLinkLabel: wideString read FXLinkLabel write FXLinkLabel;
  end;

  TXLinkArc = class (TXLinkNode)
  protected
    FActuate: TXLinkActuate;
    FArcrole: wideString;
    FXLinkFrom: wideString;
    FShow: TXLinkShow;
    FTitle: wideString;
    FTitles: TXLinkList;
    FXLinkTo: wideString;
    function createXLinkTitle(associatedObject: TObject): TXLinkTitle; virtual;
    function getShowAsWideString: wideString; virtual;
    function getActuateAsWideString: wideString; virtual;
  public
    constructor Create;
    destructor Destroy; override;
    procedure AppendContentOf(source: TXLinkArc;
                              copyObjInf: boolean); virtual;
    function AppendXLinkTitle(associatedObject: TObject): TXLinkTitle; virtual;
    function InsertXLinkTitle(index: integer;
                              associatedObject: TObject): TXLinkTitle; virtual;
    property Actuate: TXLinkActuate read FActuate write FActuate;
    property Arcrole: wideString read FArcrole write FArcrole;
    property Show: TXLinkShow read FShow write FShow;
    property ShowAsWideString: wideString read getShowAsWideString;
    property Title: wideString read FTitle write FTitle;
    property Titles: TXLinkList read FTitles;
    property XLinkFrom: wideString read FXLinkFrom write FXLinkFrom;
    property XLinkTo: wideString read FXLinkTo write FXLinkTo;
  end;

  TXLinkResource = class (TXLinkNode)
  protected
    FRole: wideString;
    FTitle: wideString;
    FXLinkLabel: wideString;
  public
    constructor Create;
    property Role: wideString read FRole write FRole;
    property Title: wideString read FTitle write FTitle;
    property XLinkLabel: wideString read FXLinkLabel write FXLinkLabel;
  end;

  TXLinkTitle = class (TXLinkNode)
  public
    constructor Create;
  end;

  TXLinkCollection = class(TXLinkList)
  protected
    function buildExtended(role, title: wideString;
                           associatedObject: TObject): TXLinkExtended; virtual;
    function buildSimple(href, role, arcrole, title: wideString;
                         show: TXLinkShow;
                         actuate: TXLinkActuate;
                         associatedObject: TObject): TXLinkExtended; virtual;
  public
    constructor Create;
    procedure AppendContentOf(source: TXLinkCollection;
                              copyObjInf: boolean); virtual;
    function AppendXLinkExtended(role, title: wideString;
                                 associatedObject: TObject): TXLinkExtended; virtual;
    function AppendXLinkSimple(href, role, arcrole, title: wideString;
                               show: TXLinkShow;
                               actuate: TXLinkActuate;
                               associatedObject: TObject): TXLinkExtended; virtual;
    function InsertXLinkExtended(index: integer;
                                 role, title: wideString;
                                 associatedObject: TObject): TXLinkExtended; virtual;
    function InsertXLinkSimple(index: integer;
                               href, role, arcrole, title: wideString;
                               show: TXLinkShow;
                               actuate: TXLinkActuate;
                               associatedObject: TObject): TXLinkExtended; virtual;
  end;

function ActuateToWideString(const value: TXLinkActuate): wideString;
function ShowToWideString(const value: TXLinkShow): wideString;
function WideStringToActuate(const s: wideString): TXLinkActuate;
function WideStringToShow(const s: wideString): TXLinkShow;

implementation

function ActuateToWideString(const value: TXLinkActuate): wideString;
begin
  case value of
    laUnspecified: result:= '';
    laOnLoad:      result:= 'onLoad';
    laOnRequest:   result:= 'onRequest';
    laOther:       result:= 'other';
    laNone:        result:= 'none';
  end;
end;

function ShowToWideString(const value: TXLinkShow): wideString;
begin
  case value of
    lsUnspecified: result:= '';
    lsNew:         result:= 'new';
    lsReplace:     result:= 'replace';
    lsEmbed:       result:= 'embed';
    lsOther:       result:= 'other';
    lsNone:        result:= 'none';
  end;
end;

function WideStringToActuate(const s: wideString): TXLinkActuate;
begin
  if s = 'onLoad' then result:= laOnLoad
  else if s = 'onRequest' then result:= laOnRequest
  else if s = 'other' then result:= laOther
  else if s = 'none' then result:= laNone
  else result:= laUnspecified;
end;

function WideStringToShow(const s: wideString): TXLinkShow;
begin
  if s = 'new' then result:= lsNew
  else if s = 'replace' then result:= lsReplace
  else if s = 'embed' then result:= lsEmbed
  else if s = 'other' then result:= lsOther
  else if s = 'none' then result:= lsNone
  else result:= lsUnspecified;
end;



// +++++++++++++++++++++++++++ TXLinkList +++++++++++++++++++++++++++
constructor TXLinkList.create(const itemTypes: TXLinkTypes);
begin
  inherited create;
  FList:= TObjectList.create;
  FItemTypes:= itemTypes;
end;

destructor TXLinkList.destroy;
begin
  FList.free;
  inherited destroy;
end;

function TXLinkList.getItems(index: integer): TXLinkNode;
begin
  result:= TXLinkNode(FList.Items[index]);
end;

function TXLinkList.getLength: integer;
begin
  result:= FList.count;
end;

function TXLinkList.getOwnsXLinks: boolean;
begin
  result:= FList.OwnsObjects;
end;

procedure TXLinkList.setOwnsXLinks(value: boolean);
begin
  FList.OwnsObjects:= value;
end;

function TXLinkList.Add(item: TXLinkNode): integer;
begin
  if not assigned(item)
    then raise EXLink_Syntax_Err.Create('XLink: Syntax error');
  if not (item.XLinkType in FItemTypes)
    then raise EXLink_Hierarchy_Request_Err.create('XLink: Hierarchy request error.');
  if assigned(item.Owner)
    then raise EXLink_Inuse_Node_Err.create('XLink: Inuse node error.');
  result:= FList.Add(item);
  item.FOwner:= self;
end;

function TXLinkList.AddCopyOf(item: TXLinkNode;
                              deep: boolean): TXLinkNode;
var
  newItem: TXLinkNode;
  i: integer;
begin
  if not assigned(item)
    then raise EXLink_Syntax_Err.Create('XLink: Syntax error: TXLinkNode not specified');
  if not (item.XLinkType in FItemTypes)
    then raise EXLink_Hierarchy_Request_Err.create('XLink: Hierarchy request error.');
  result:= nil;
  case item.XLinkType of
    ltNone: raise EXLink_Syntax_Err.Create('XLink: Syntax error: TXLinkNode of type ltNone cannot be copied');
    ltExtended: begin  // xxx better define and use assign
      newItem:= TXLinkExtended.Create;
      TXLinkExtended(newItem).role:= TXLinkExtended(item).role;
      TXLinkExtended(newItem).title:= TXLinkExtended(item).title;
      add(newItem);
      result:= newItem;
      if deep then begin
        with TXLinkExtended(item).resources do
          for i:= 0 to pred(length) do
            TXLinkExtended(newItem).resources.AddCopyOf(items[i],true);
        with TXLinkExtended(item).locators do
          for i:= 0 to pred(length) do
            TXLinkExtended(newItem).locators.AddCopyOf(items[i],true);
        with TXLinkExtended(item).arcs do
          for i:= 0 to pred(length) do
            TXLinkExtended(newItem).arcs.AddCopyOf(items[i],true);
        with TXLinkExtended(item).titles do
          for i:= 0 to pred(length) do
            TXLinkExtended(newItem).titles.AddCopyOf(items[i],true);
      end;
    end;
    ltLocator: begin
      newItem:= TXLinkLocator.Create;
      TXLinkLocator(newItem).href:= TXLinkLocator(item).href;
      TXLinkLocator(newItem).role:= TXLinkLocator(item).role;
      TXLinkLocator(newItem).title:= TXLinkLocator(item).title;
      TXLinkLocator(newItem).XLinkLabel:= TXLinkLocator(item).XLinkLabel;
      add(newItem);
      result:= newItem;
      if deep then begin
        with TXLinkLocator(item).titles do
          for i:= 0 to pred(length) do
            TXLinkLocator(newItem).titles.AddCopyOf(items[i],true);
      end;
    end;
    ltArc: begin
      newItem:= TXLinkArc.Create;
      TXLinkArc(newItem).arcrole:= TXLinkArc(item).arcrole;
      TXLinkArc(newItem).title:= TXLinkArc(item).title;
      TXLinkArc(newItem).show:= TXLinkArc(item).show;
      TXLinkArc(newItem).actuate:= TXLinkArc(item).actuate;
      TXLinkArc(newItem).XLinkFrom:= TXLinkArc(item).XLinkFrom;
      TXLinkArc(newItem).XLinkTo:= TXLinkArc(item).XLinkTo;
      add(newItem);
      result:= newItem;
      if deep then begin
        with TXLinkArc(item).titles do
          for i:= 0 to pred(length) do
            TXLinkArc(newItem).titles.AddCopyOf(items[i],true);
      end;
    end;
    ltResource: begin
      newItem:= TXLinkResource.Create;
      TXLinkResource(newItem).role:= TXLinkResource(item).role;
      TXLinkResource(newItem).title:= TXLinkResource(item).title;
      TXLinkResource(newItem).XLinkLabel:= TXLinkResource(item).XLinkLabel;
      add(newItem);
      result:= newItem;
    end;
    ltTitle: begin
      newItem:= TXLinkTitle.Create;
      add(newItem);
      result:= newItem;
    end;
  end;
end;

function TXLinkList.Extract(item: TXLinkNode): TXLinkNode;
begin
  if not assigned(item)
    then raise EXLink_Syntax_Err.Create('XLink: Syntax error');
  result:= TXLinkNode(FList.Extract(item));
end;

function TXLinkList.FindInstanceOf(AClass: TClass; AExact: boolean = true; AStartAt: integer = 0): integer;
begin
  result:= FList.FindInstanceOf(AClass,AExact,AStartAt);
end;

procedure TXLinkList.Insert(index: integer; item: TXLinkNode);
begin
  if not assigned(item)
    then raise EXLink_Syntax_Err.Create('XLink: Syntax error');
  if not (item.XLinkType in FItemTypes)
    then raise EXLink_Hierarchy_Request_Err.create('XLink: Hierarchy request error.');
  if assigned(item.Owner)
    then raise EXLink_Inuse_Node_Err.create('XLink: Inuse node error.');
  FList.Insert(index,item);
  item.FOwner:= self;
end;

procedure TXLinkList.Clear;
begin
  FList.Clear;
end;

procedure TXLinkList.Delete(index: integer);
begin
  FList.delete(index);
end;

procedure TXLinkList.Exchange(index1, index2: integer);
begin
  FList.Exchange(index1,index2);
end;

function TXLinkList.First: TXLinkNode;
begin
  result:= TXLinkNode(FList.first);
end;

function TXLinkList.IndexOf(item: TXLinkNode): integer;
begin
  result:= FList.IndexOf(item);
end;

function TXLinkList.Last: TXLinkNode;
begin
  result:= TXLinkNode(FList.last);
end;

procedure TXLinkList.Move(curIndex, newIndex: integer);
begin
  FList.move(curIndex,newIndex);
end;

function TXLinkList.Remove(item: TXLinkNode): integer;
begin
  if not assigned(item)
    then raise EXLink_Syntax_Err.Create('XLink: Syntax error');
  result:= FList.Remove(item);
end;



// +++++++++++++++++++++++++++ TXLinkNode +++++++++++++++++++++++++++
constructor TXlinkNode.create;
begin
  inherited create;
  FAssociatedObject:= nil;
  FOwner:= nil;
  FXLinkType:= ltNone;
end;

function TXLinkNode.getTypeAsWideString: wideString;
begin
  case FXLinkType of
    ltNone:     result:= 'none';
    ltExtended: result:= 'extended';
    ltLocator:  result:= 'locator';
    ltArc:      result:= 'arc';
    ltResource: result:= 'resource';
    ltTitle:    result:= 'title';
  end;
end;



// +++++++++++++++++++++++++ TXLinkExtended +++++++++++++++++++++++++
constructor TXLinkExtended.Create;
begin
  inherited create;
  FArcs:= TXLinkList.create([ltArc]);
  FLocators:= TXLinkList.create([ltLocator]);
  FResources:= TXLinkList.create([ltResource]);
  FRole:= '';
  FTitle:= '';
  FTitles:= TXLinkList.create([ltTitle]);
  FXLinkType:= ltExtended;
end;

destructor TXLinkExtended.Destroy;
begin
  FArcs.free;
  FLocators.free;
  FResources.free;
  FTitles.free;
  inherited destroy;
end;

function TXLinkExtended.createXLinkArc(arcrole, title: wideString;
                                       show: TXLinkShow;
                                       actuate: TXLinkActuate;
                                       XLinkFrom, XLinkTo: wideString;
                                       associatedObject: TObject): TXLinkArc;
begin
  result:= TXLinkArc.Create;
  result.arcrole:= arcrole;
  result.title:= title;
  result.show:= show;
  result.actuate:= actuate;
  result.XLinkFrom:= XLinkFrom;
  result.XLinkTo:= XLinkTo;
  result.AssociatedObject:= associatedObject;
end;

function TXLinkExtended.createXLinkLocator(href, role, title, XLinklabel: wideString;
                                           associatedObject: TObject): TXLinkLocator;
begin
  result:= TXLinkLocator.Create;
  result.href:= href;
  result.role:= role;
  result.title:= title;
  result.Xlinklabel:= XLinkLabel;
  result.AssociatedObject:= associatedObject;
end;

function TXLinkExtended.createXLinkResource(role, title, XLinkLabel: wideString;
                                            associatedObject: TObject): TXLinkResource;
begin
  result:= TXLinkResource.Create;
  result.role:= role;
  result.title:= title;
  result.Xlinklabel:= XLinkLabel;
  result.AssociatedObject:= associatedObject;
end;

function TXLinkExtended.createXLinkTitle(associatedObject: TObject): TXLinkTitle;
begin
  result:= TXLinkTitle.Create;
  result.AssociatedObject:= associatedObject;
end;

procedure TXLinkExtended.AppendContentOf(source: TXLinkExtended;
                                         copyObjInf: boolean);
var
  i: integer;
  obj: TObject;
  newArc: TXLinkArc;
  newLocator: TXLinkLocator;
begin
  if (source = self) or not assigned(source) then exit;
  obj:= nil;
  for i:= 0 to pred(source.Arcs.Length) do begin
    with TXLinkArc(source.Arcs.Items[i]) do begin
      if copyObjInf then obj:= AssociatedObject;
      newArc:= TXLinkArc(self.AppendXLinkArc(arcrole,title,show,actuate,XLinkFrom,XLinkTo,obj));
      newArc.AppendContentOf(TXLinkArc(source.Arcs.Items[i]),copyObjInf);
    end;
  end; {for ...}
  for i:= 0 to pred(source.Locators.Length) do begin
    with TXLinkLocator(source.Locators.Items[i]) do begin
      if copyObjInf then obj:= AssociatedObject;
      newLocator:= TXLinkLocator(self.AppendXLinkLocator(href,role,title,XLinkLabel,obj));
      newLocator.AppendContentOf(TXLinkLocator(source.Locators.Items[i]),copyObjInf);
    end;
  end; {for ...}
  for i:= 0 to pred(source.Resources.Length) do begin
    with TXLinkResource(source.Resources.Items[i]) do begin
      if copyObjInf then obj:= AssociatedObject;
      self.AppendXLinkResource(role,title,XLinkLabel,obj);
    end;
  end; {for ...}
  for i:= 0 to pred(source.Titles.Length) do begin
    with TXLinkTitle(source.Titles.Items[i]) do begin
      if copyObjInf then obj:= AssociatedObject;
      self.AppendXLinkTitle(obj);
    end;
  end; {for ...}
end;

function TXLinkExtended.AppendXLinkArc(arcrole, title: wideString;
                                       show: TXLinkShow;
                                       actuate: TXLinkActuate;
                                       XLinkFrom, XLinkTo: wideString;
                                       associatedObject: TObject): TXLinkArc;
begin
  result:= createXLinkArc(arcrole,title,show,actuate,XLinkFrom,XLinkTo,associatedObject);
  FArcs.Add(result);
end;

function TXLinkExtended.AppendXLinkLocator(href, role, title, XLinklabel: wideString;
                                           associatedObject: TObject): TXLinkLocator;
begin
  result:= createXLinkLocator(href,role,title,XLinklabel,associatedObject);
  FLocators.Add(result);
end;

function TXLinkExtended.AppendXLinkResource(role, title, XLinkLabel: wideString;
                                            associatedObject: TObject): TXLinkResource;
begin
  result:= createXLinkResource(role,title,XLinklabel,associatedObject);
  FResources.Add(result);
end;

function TXLinkExtended.AppendXLinkTitle(associatedObject: TObject): TXLinkTitle;
begin
  result:= createXLinkTitle(associatedObject);
  FTitles.Add(result);
end;

function TXLinkExtended.InsertXLinkArc(index: integer;
                                       arcrole, title: wideString;
                                       show: TXLinkShow;
                                       actuate: TXLinkActuate;
                                       XLinkFrom, XLinkTo: wideString;
                                       associatedObject: TObject): TXLinkArc;
begin
  result:= self.createXLinkArc(arcrole,title,show,actuate,XLinkFrom,XLinkTo,associatedObject);
  FTitles.insert(index,result);
end;

function TXLinkExtended.InsertXLinkLocator(index: integer;
                                           href, role, title, XLinklabel: wideString;
                                           associatedObject: TObject): TXLinkLocator;
begin
  result:= self.createXLinkLocator(href,role,title,XLinklabel,associatedObject);
  FTitles.insert(index,result);
end;

function TXLinkExtended.InsertXLinkResource(index: integer;
                                            role, title, XLinkLabel: wideString;
                                            associatedObject: TObject): TXLinkResource;
begin
  result:= self.createXLinkResource(role,title,XLinklabel,associatedObject);
  FTitles.insert(index,result);
end;

function TXLinkExtended.InsertXLinkTitle(index: integer;
                                         associatedObject: TObject): TXLinkTitle;
begin
  result:= self.createXLinkTitle(associatedObject);
  FTitles.insert(index,result);
end;



// +++++++++++++++++++++++++ TXLinkLocator ++++++++++++++++++++++++++
constructor TXLinkLocator.Create;
begin
  inherited create;
  FHref:= '';
  FRole:= '';
  FTitle:= '';
  FTitles:= TXLinkList.create([ltTitle]);
  FXLinkLabel:= '';
  FXLinkType:= ltLocator;
end;

destructor TXLinkLocator.Destroy;
begin
  FTitles.free;
  inherited destroy;
end;

function TXLinkLocator.createXLinkTitle(associatedObject: TObject): TXLinkTitle;
begin
  result:= TXLinkTitle.Create;
  result.AssociatedObject:= associatedObject;
end;

procedure TXLinkLocator.AppendContentOf(source: TXLinkLocator;
                                        copyObjInf: boolean);
var
  i: integer;
  obj: TObject;
begin
  if (source = self) or not assigned(source) then exit;
  obj:= nil;
  for i:= 0 to pred(source.Titles.Length) do begin
    with TXLinkTitle(source.Titles.Items[i]) do begin
      if copyObjInf then obj:= AssociatedObject;
      self.AppendXLinkTitle(obj);
    end;
  end; {for ...}
end;

function TXLinkLocator.AppendXLinkTitle(associatedObject: TObject): TXLinkTitle;
begin
  result:= createXLinkTitle(associatedObject);
  FTitles.Add(result);
end;

function TXLinkLocator.InsertXLinkTitle(index: integer;
                                       associatedObject: TObject): TXLinkTitle;
begin
  result:= self.createXLinkTitle(associatedObject);
  FTitles.insert(index,result);
end;



// +++++++++++++++++++++++++++ TXLinkArc ++++++++++++++++++++++++++++
constructor TXLinkArc.Create;
begin
  inherited create;
  FActuate:= laUnspecified;
  FArcrole:= '';
  FShow:= lsUnspecified;
  FTitle:= '';
  FTitles:= TXLinkList.create([ltTitle]);
  FXLinkFrom:= '';
  FXLinkTo:= '';
  FXLinkType:= ltArc;
end;

destructor TXLinkArc.Destroy;
begin
  FTitles.free;
  inherited destroy;
end;

function TXLinkArc.createXLinkTitle(associatedObject: TObject): TXLinkTitle;
begin
  result:= TXLinkTitle.Create;
  result.AssociatedObject:= associatedObject;
end;

procedure TXLinkArc.AppendContentOf(source: TXLinkArc;
                                    copyObjInf: boolean);
var
  i: integer;
  obj: TObject;
begin
  if (source = self) or not assigned(source) then exit;
  obj:= nil;
  for i:= 0 to pred(source.Titles.Length) do begin
    with TXLinkTitle(source.Titles.Items[i]) do begin
      if copyObjInf then obj:= AssociatedObject;
      self.AppendXLinkTitle(obj);
    end;
  end; {for ...}
end;

function TXLinkArc.AppendXLinkTitle(associatedObject: TObject): TXLinkTitle;
begin
  result:= createXLinkTitle(associatedObject);
  FTitles.Add(result);
end;

function TXLinkArc.InsertXLinkTitle(index: integer;
                                    associatedObject: TObject): TXLinkTitle;
begin
  result:= self.createXLinkTitle(associatedObject);
  FTitles.insert(index,result);
end;

function TXLinkArc.getActuateAsWideString: wideString;
begin
  result:= ActuateToWideString(FActuate);
end;

function TXLinkArc.getShowAsWideString: wideString;
begin
  result:= ShowToWideString(FShow);
end;



// +++++++++++++++++++++++++ TXLinkResource +++++++++++++++++++++++++
constructor TXLinkResource.Create;
begin
  inherited create;
  FRole:= '';
  FTitle:= '';
  FXLinkLabel:= '';
  FXLinkType:= ltResource;
end;



// ++++++++++++++++++++++++++ TXLinkTitle +++++++++++++++++++++++++++
constructor TXLinkTitle.Create;
begin
  inherited create;
  FXLinkType:= ltTitle;
end;



// ++++++++++++++++++++++++ TXLinkCollection ++++++++++++++++++++++++
constructor TXLinkCollection.create;
begin
  inherited create([ltExtended]);
end;

function TXLinkCollection.buildExtended(role, title: wideString;
                                        associatedObject: TObject): TXLinkExtended;
begin
  result:= TXLinkExtended.Create;
  result.Role:= role;
  result.Title:= title;
  result.AssociatedObject:= associatedObject;
end;

function TXLinkCollection.buildSimple(href, role, arcrole, title: wideString;
                                      show: TXLinkShow;
                                      actuate: TXLinkActuate;
                                      associatedObject: TObject): TXLinkExtended;
begin
  result:= TXLinkExtended.Create;
  result.AppendXLinkResource('','','local',associatedObject);
  if href = '' then begin
    result.AppendXLinkArc(arcrole,'',show,actuate,'local','',associatedObject);
  end else begin
    result.AppendXLinkLocator(href,role,title,'remote',associatedObject);
    result.AppendXLinkArc(arcrole,'',show,actuate,'local','remote',associatedObject);
  end;
end;

procedure TXLinkCollection.AppendContentOf(source: TXLinkCollection;
                                           copyObjInf: boolean);
var
  i: integer;
  obj: TObject;
  newExtended: TXLinkExtended;
begin
  if (source = self) or not assigned(source) then exit;
  obj:= nil;
  for i:= 0 to pred(source.Length) do begin
    with TXLinkExtended(source.Items[i]) do begin
      if copyObjInf then obj:= AssociatedObject;
      newExtended:= TXLinkExtended(self.AppendXLinkExtended(role,title,obj));
      newExtended.AppendContentOf(TXLinkExtended(source.Items[i]),copyObjInf);
    end;
  end; {for ...}
end;

function TXLinkCollection.AppendXLinkExtended(role, title: wideString;
                                              associatedObject: TObject): TXLinkExtended;
begin
  result:= buildExtended(role,title,associatedObject);
  Add(result);
end;

function TXLinkCollection.AppendXLinkSimple(href, role, arcrole, title: wideString;
                                            show: TXLinkShow;
                                            actuate: TXLinkActuate;
                                            associatedObject: TObject): TXLinkExtended;
begin
  result:= buildSimple(href,role,arcrole,title,show,actuate,associatedObject);
  Add(result);
end;

function TXLinkCollection.InsertXLinkExtended(index: integer;
                                              role, title: wideString;
                                              associatedObject: TObject): TXLinkExtended;
begin
  result:= self.buildExtended(role,title,associatedObject);
  insert(index,result);
end;

function TXLinkCollection.InsertXLinkSimple(index: integer;
                                            href, role, arcrole, title: wideString;
                                            show: TXLinkShow;
                                            actuate: TXLinkActuate;
                                            associatedObject: TObject): TXLinkExtended;
begin
  result:= self.buildSimple(href,role,arcrole,title,show,actuate,associatedObject);
  insert(index,result);
end;

end.
